import json
import flask_transmute
from flask import Flask, Blueprint
from flask_transmute.swagger import SWAGGER_ATTR_NAME


def test_blueprint_only_app():
    """
    A flask app that only has transmute routes
    in blueprints should be valide.
    """
    app = Flask(__name__)
    blueprint = Blueprint("foo", __name__)

    # this also unit tests empty url prefix blueprints.
    @flask_transmute.route(blueprint, paths=["/foo"])
    def foo():
        return None

    app.register_blueprint(blueprint)
    flask_transmute.add_swagger(app, "/swagger.json", "/swagger")

    test_client = app.test_client()
    body = json.loads(test_client.get("/swagger.json").data)
    assert "/foo" in body["paths"]



import json
from flask import Response, Blueprint
from .legacy import Swagger

from transmute_core.swagger import (
    generate_swagger_html,
    get_swagger_static_root,
    SwaggerSpec
)

STATIC_ROOT = "/_swagger/static"
SWAGGER_ATTR_NAME = "_transmute_swagger"


def add_swagger(app, json_route, html_route):
    """
    a convenience method for both adding a swagger.json route,
    as well as adding a page showing the html documentation
    """
    app.route(json_route)(create_swagger_json_handler(app))
    add_swagger_api_route(app, html_route, json_route)


def add_swagger_api_route(app, target_route, swagger_json_route):
    """
    mount a swagger statics page.

    app: the flask app object
    target_route: the path to mount the statics page.
    swagger_json_route: the path where the swagger json definitions is
        expected to be.
    """
    static_root = get_swagger_static_root()
    swagger_body = generate_swagger_html(
        STATIC_ROOT, swagger_json_route
    ).encode("utf-8")

    def swagger_ui():
        return Response(swagger_body, content_type="text/html")

    blueprint = Blueprint('swagger', __name__,
                          static_url_path=STATIC_ROOT,
                          static_folder=static_root)
    app.route(target_route)(swagger_ui)
    app.register_blueprint(blueprint)


def create_swagger_json_handler(app, **kwargs):
    """
    Create a handler that returns the swagger definition
    for an application.

    This method assumes the application is using the
    TransmuteUrlDispatcher as the router.
    """

    spec = getattr(app, SWAGGER_ATTR_NAME, SwaggerSpec())
    _add_blueprint_specs(app, spec)
    spec_dict = spec.swagger_definition(**kwargs)
    encoded_spec = json.dumps(spec_dict).encode("UTF-8")

    def swagger():
        return Response(
            encoded_spec,
            # we allow CORS, so this can be requested at swagger.io
            headers={
                "Access-Control-Allow-Origin": "*"
            },
            content_type="application/json",
        )

    return swagger


def _add_blueprint_specs(app, root_spec):
    for name, blueprint in app.blueprints.items():
        if hasattr(blueprint, SWAGGER_ATTR_NAME):
            spec = getattr(blueprint, SWAGGER_ATTR_NAME)
            for path, path_item in spec.paths.items():
                if blueprint.url_prefix:
                    path = blueprint.url_prefix + path
                root_spec.add_path(path, path_item)



import json


def test_happy_path(test_app):
    r = test_app.get("/multiply?left=3&right=3")
    assert json.loads(r.data.decode()) == 9


def test_complex(test_app):
    r = test_app.post("/complex/3",
                      data=json.dumps({"body": "1"}),
                      headers={
                          "header": "2",
                          "content-type": "application/json"
                      })
    assert json.loads(r.data.decode()) == "1:2:3"


def test_api_exception(test_app):
    r = test_app.get("/exception")
    assert r.status_code == 400
    resp = json.loads(r.data.decode())
    assert resp["code"] == 400
    assert resp["success"] is False


def test_swagger(test_app):
    r = test_app.get("/swagger.json")
    swagger = json.loads(r.data.decode())
    assert swagger["info"] == {
        'version': '1.0',
        'title': 'example'
    }
    assert '/multiply' in swagger['paths']
    assert '/exception' in swagger['paths']
    # test blueprint is documented as well
    assert '/blueprint/foo' in swagger['paths']


def test_swagger_html(test_app):
    r = test_app.get("/swagger")
    assert "/swagger.json" in r.data.decode()
    assert r.status_code == 200



from flask_transmute import route, annotate
import flask_transmute
from flask import Flask, Blueprint

app = Flask(__name__)


@flask_transmute.route(app, paths='/multiply')
@annotate({"left": int, "right": int, "return": int})
def multiply(left, right):
    return left * right


@route(app, paths='/exception')
def exception():
    raise flask_transmute.APIException("api error!")


@flask_transmute.route(app,
                       paths='/complex/{path}',
                       methods=["POST"],
                       body_parameters=["body"],
                       header_parameters=["header"])
@annotate({"body": str, "header": str, "path": str, "return": str})
def complex(body, header, path):
    return body + ":" + header + ":" + path

blueprint = Blueprint('blueprint', __name__, url_prefix="/blueprint")


@flask_transmute.route(blueprint, paths='/foo')
@annotate({"return": bool})
def foo():
    return True

app.register_blueprint(blueprint)

# finally, you can add a swagger json and a documentation page by:
flask_transmute.add_swagger(app, "/swagger.json", "/swagger")



LEGACY_USAGE_ERROR = """
looks like you are attempting to
use a function in the pre-1.0
flask-transmute, which is no longer
supported.

The last pre-1.0 release is 0.2.15. You
can pin to that, or upgrade to the new
syntax and approach:

http://flask-transmute.readthedocs.io/en/latest/legacy/migrating.html
http://transmute-core.readthedocs.io/en/latest/
""".strip()


def FlaskRouteSet(*args, **kwargs):
    raise NotImplementedError(LEGACY_USAGE_ERROR)


def Swagger(*args, **kwargs):
    raise NotImplementedError(LEGACY_USAGE_ERROR)



from transmute_core import (
    describe, default_context, TransmuteFunction
)
from transmute_core.swagger import SwaggerSpec
from .handler import create_routes_and_handler
from .swagger import SWAGGER_ATTR_NAME


def route(app_or_blueprint, context=default_context, **kwargs):
    """ attach a transmute route. """
    def decorator(fn):
        fn = describe(**kwargs)(fn)
        transmute_func = TransmuteFunction(fn)
        routes, handler = create_routes_and_handler(transmute_func, context)
        for r in routes:
            # push swagger info.
            if not hasattr(app_or_blueprint, SWAGGER_ATTR_NAME):
                setattr(app_or_blueprint, SWAGGER_ATTR_NAME, SwaggerSpec())
            swagger_obj = getattr(app_or_blueprint, SWAGGER_ATTR_NAME)
            swagger_obj.add_func(transmute_func, context)
            app_or_blueprint.route(r, methods=transmute_func.methods)(handler)
        return handler
    return decorator



import pytest
from flask_transmute import FlaskRouteSet
from flask_transmute.swagger import Swagger


@pytest.mark.parametrize("f", [
    FlaskRouteSet, Swagger
])
def test_legacy_error(f):
    """ invoking legacy function should result in an exception. """
    with pytest.raises(NotImplementedError):
        f()



from functools import wraps
from flask import request, Response
from transmute_core import (
    ParamExtractor, NoArgument
)


def create_routes_and_handler(transmute_func, context):

    @wraps(transmute_func.raw_func)
    def handler(*args, **kwargs):
        exc, result = None, None
        try:
            args, kwargs = _param_instance.extract_params(
                context, transmute_func, request.content_type,
            )
            result = transmute_func(*args, **kwargs)
        except Exception as e:
            exc = e
        response = transmute_func.process_result(
            context, result, exc, request.content_type
        )
        return Response(
            response["body"],
            status=response["code"],
            mimetype=response["content-type"]
        )
    return (
        _convert_paths_to_flask(transmute_func.paths),
        handler
    )


def _convert_paths_to_flask(transmute_paths):
    """ flask has it's own route syntax, so we convert it. """
    paths = []
    for p in transmute_paths:
        paths.append(p.replace("{", "<").replace("}", ">"))
    return paths


class ParamExtractorFlask(ParamExtractor):

    def _get_framework_args(self):
        return {}

    @property
    def body(self):
        return request.get_data()

    @staticmethod
    def _query_argument(key, is_list):
        if key not in request.args:
            return NoArgument
        if is_list:
            return request.args.getlist(key)
        else:
            return request.args[key]

    @staticmethod
    def _header_argument(key):
        return request.headers.get(key, NoArgument)

    @staticmethod
    def _path_argument(key):
        return request.view_args.get(key, NoArgument)

_param_instance = ParamExtractorFlask()
